/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     coded in ECMAScript 5 - JavaScript 5  (ES5, ECMAScript 2009)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


//  xxxxxxxxxxxxxxx     redundancy-web variable declarations  xxxxxxxxxxxxxxxx


// These variables below can be directly modified by the user on the options form: 

var item_cnt= 3;       // The number of stimuli in the task

var stype = '0';      //  stype => Stimulus type: "0" => Alphanumeric characters; "1"  => Images (selected by the user)

var alfa_type = '0';  // Type of Alphanumeric characters => "0" = Digits; "1" = Letters-lowercase; "2" = Letters-UPPERCASE; "3" = All digits and letters

var alfa_cnt = 1;  // The number of alphanumeric characters in the task

var font_family = 'Georgia, serif';  // default canvas ctx.fontFamily (see drawText function)

var mtype = '0';   // Type of stimulus masking

var stripe_cnt= 5;  // Number of stripes that mask the stimulus (raster). Enabled only for stripe masking methods

var slevel= 21;      // The initial level of masking each stimulus. Lower value = task is easier. Higher value = the task is more difficult
   
var demo = false;     // if (demo == true) => Show correct solution (not for Stimulus type = images)

var feedback = true;    // Additional information will be displayed at the bottom of the page (true) or not (false)

var mask_clr= '#ffffff';  // Mask and background (canvas) color. Enabled for image stimuli only

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other redundancy-web variables and constants:

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_task = false;  // task in progress or not => for html: body onresize="pageResize()"

var item = 0;   // current item      

var clevel;  		// Current Level for stimulus masking

var stim = [];  // Stimulus displayed to the user (alphanumeric characters)

var ans = [];  	//  user (client) answer for each item 

var percent = []; // Percentage of invisible (mask) pixels for each item = 100 * number of masked pixels / total number of pixels  

var per_arr = [ 0, 0.01, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95, 0.99 ];   // Percentage of black pixels to be set to invisible for The initial level of masking each stimulus. index 0..21

var font_size= 24;   // global fontSize (px) for drawText() function

var bounds={};      // image stimulus bounds on canvas 


//  xxxxxxxxxxxxxxx     redundancy-web functions  xxxxxxxxxxxxxxxx


// app init, init page layout, [center panels], and show start panel.. 
function init() 
{      
  if (isStorage() && storageAvailable('localStorage'))  
  {
    document.getElementById("store_group").disabled= false;
  }
  else 
  {
    document.getElementById("store_msg").textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline).";
    document.getElementById("store_group").disabled= true;  
    document.getElementById("clr_store").disabled= true;     // IE11 bug  => childrens are not auto disabled
    document.getElementById("store").disabled= true;          // IE11 bug => childrens are not auto disabled
  }    
	
	if ( !readerExists() ) {
	  document.getElementById('stype').disabled= true;
		document.getElementById('no_reader').style.display= 'block';		
	}	
	
  document.getElementById('answer').onkeyup = answerKeyUp;	
  
	setIns();
	
	loadFont();
	
	showStart();  
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== "undefined") {
     return true;               // Code for localStorage/sessionStorage.     
  } 
  else {
    return false;           // code for no browser Web Storage support
  }  
}


// necessarily because of IE11 and Edge offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// Checks if the client's web browser supports this: FileReader object, Blob object, FileReader.readAsDataURL() 
function readerExists() 
{
	try
	{
		var img = document.getElementById('img_tmp');
		// create Blob from a typed array and strings: https://javascript.info/blob
		var hello = new Uint8Array([72, 101, 108, 108, 111]); // "Hello" in binary form
		var blob = new Blob([hello, ' ', 'world'], {type: 'text/plain'});		
		var reader = new FileReader();
		if (blob) {
			reader.readAsDataURL(blob);
			return true;
		}
	}	 
  catch(e) { 
    return false;
  }
	return false;
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{
	if (is_task) { 
		setPage();
	}
}	


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	var arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	var sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (var i= 0; i!= len; i++) 
	{  
		opt = document.createElement("option");
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex=1; // Georgia, serif
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[1];
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById("hide_option").style.display ="block";
  document.getElementById("options").style.display= "block";
  document.getElementById("start").classList.remove("child");  // don't work in <= IE 9    
  document.getElementById("start_div").style.display="none";  // .disabled = true;   
  document.getElementById("show_option").style.display ="none";  
  document.getElementById("hide_option").style="text-align:center";  // bug? auto-remove style when element.style.display="none"
  document.getElementById("instructions").style.display ="none";     
  if ( document.getElementById("store_group").disabled === false && localStorage.getItem("redundancy-web-v1.0")) 
  {
    var cnt= JSON.parse(localStorage.getItem("redundancy-web-v1.0")).length;                                                        
    document.getElementById("store_cnt").textContent=  "You have a total of " + cnt  + " redundancy-web results stored in your web browser";
  }
  dirty= false;
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById("options").style.display = "none"; 
  document.getElementById("start").classList.add("child");   
  document.getElementById("show_option").style.display = "block";   
  document.getElementById("start_div").style.display = "block";        //.removeAttribute("disabled");  
  document.getElementById("hide_option").style.display = "none";   
  document.getElementById("start_div").style="text-align:center";  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById("show_option").style="text-align:center";   // bug?
  document.getElementById("instructions").style.display = "block";
  refMsg(); 
  setIns();	
}


// set instructions text
function setIns()  
{
  document.getElementById("instructions").style.display= "block";	
	if (stype === "1")   {
	  document.getElementById("ins_1").style.display= 'block';
		document.getElementById("ins_0").style.display= 'none';
		document.getElementById("ins_1_header").style.display= document.getElementById("ins_header").checked ? 'block' : 'none';
	}
  else  {
		document.getElementById("ins_0").style.display= 'block';
		document.getElementById("ins_1").style.display= 'none';
	}			
}	

       
// refresh Options (update *_msg label) => give user short option HELP
function refOption(e)  
{   
  dirty=true;       
  var my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';    
           
	switch (my_id)
	{
		
		case 'item_cnt':
			itemsValid();
			break;
		
		case 'stype':
		  if (e.value === '1')  {
				msg.textContent= 'Images will be displayed';
				document.getElementById('alfa_opt').style.display= 'none';        // document.getElementById("alfa_opt").disabled= true;
				document.getElementById('images_opt').style.display= 'block';   // 	document.getElementById("images_opt").disabled= false;
				document.getElementById('demo').disabled= true;
				//document.getElementById('mask_clr_msg').textContent= 'Mask color for image stimuli. Adjust it to your image stimuli.';
			}
			else  {
				msg.textContent= 'Random alphanumeric characters will be displayed'; 
			  document.getElementById('images_opt').style.display= 'none';  // 	document.getElementById("images_opt").disabled= true;
				document.getElementById('alfa_opt').style.display= 'block';    // document.getElementById("alfa_opt").disabled= false;
				document.getElementById('demo').disabled= false;
				//document.getElementById('mask_clr_msg').textContent= 'Disabled for alphanumeric characters (default = white). Enabled for image stimuli only.';
			}	
      itemsValid();
      setMask();			
			break;
		
		case 'alfa_cnt':
			if (e.validity.valid)  {           
				msg.removeAttribute("class");       
				msg.textContent= 'The number of alphanumeric characters in the task';       //msg.textContent=  e.value; 
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 
			break;
		
    case 'mtype': 		
			if (e.value === '2' || e.value === '3')  {							
				document.getElementById('stripe_cnt').disabled= true;
				document.getElementById('stripe_cnt_msg').textContent= 'Not enabled for the selected stimulus masking type';
			}
			else  {
	 			document.getElementById('stripe_cnt').disabled= false;
				document.getElementById('stripe_cnt_msg').textContent= 'Number of stripes that mask the stimulus (raster).';
			}
      switch (e.value)  
			{
				case "0":
					msg.textContent= 'The stimuli will be masked (cut off) in the form of horizontal stripes.';
					break;
				case "1":
					msg.textContent= 'The stimuli will be masked (cut off) in the form of vertical stripes.';
					break;
				case "2":
					msg.textContent= 'The stimuli will be masked (cut off) in one piece - horizontally.';
					break;
				case "3":
					msg.textContent= 'The stimuli will be masked (cut off) in one piece - vertically.';
					break;	
				case "4": 
					msg.textContent= 'The stimuli will be masked in the form of grid. This type of mask is not difficult because mask can not fully cover the stimulus (depending on the geometry of the image). This is only enabled for Stimulus type: images';
					break;
				case "5": 
					msg.textContent= 'The stimuli will be masked in the form of a rectangles (inverse grid effect). This is only enabled for Stimulus type: images';
					break;  
				case "6": 
					msg.textContent= 'Circles mask. The stimuli will be masked in the shape of a circles. This type of mask is usually the lightest. Depends on the geometry of the loaded image. This is only enabled for Stimulus type: images';
					break;					
			}				
			break;  // outer switch
		
		case 'stripe_cnt':
			if (e.validity.valid)  {         
				msg.removeAttribute("class");       
				msg.textContent= 'Number of stripes that mask the stimulus (raster).';       //msg.textContent=  e.value; 
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 
			break;
					
		case 'slevel':  
			if (e.validity.valid)  {           
				msg.removeAttribute("class");       				   
				msg.textContent= 'The initial level of masking each stimulus. Lower value = task is easier. Higher value = the task is more difficult';    
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 					
			break; 

		case 'demo':   
			msg.textContent= e.checked ? 'Yes' : 'No';    
			break;
		
		case 'feedback':   
			msg.textContent= e.checked ? 'Additional information will be displayed at the bottom of the page.' : 'No additional information is displayed at the bottom of the page.';    
			break;
						 
		case 'mask_clr': 
			if (colorValid(e.value))   {           
				msg.removeAttribute('class');       
				msg.textContent= (document.getElementById('stype').value === "1") ? 'Mask and background color for image stimuli. Adjust it to your image stimuli.' : 'Disabled for alphanumeric characters (default = white). Enabled for image stimuli only.';  
			}
			else  {
				msg.textContent= 'The mask color is invalid.';
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 
			break; 
		 
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// check if items count is valid or not
function itemsValid()  
{
  var e= document.getElementById('item_cnt'), msg= document.getElementById('item_cnt_msg'), s = 'Values within this range are allowed: ', s2= ' with step= '; 
	var stim_type= document.getElementById('stype').value;  // stimulus type
	var images= document.getElementById('preview').getElementsByTagName('img');  
	var img_cnt = (images && images.length) ? images.length : 0;
	if ( ( e.validity.valid === false ) ||
       (( stim_type === '1' && parseInt(e.value, 10) > img_cnt ) ) )  {
		msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
		if ( stim_type === '1') {
			msg.textContent+= '. For Stimulus type Images: Item count can not be greater then images count. Images count=' + img_cnt; 
    }		
		msg.setAttribute('class', 'error');    // msg.className = 'error';
		return;		
	}	
	msg.removeAttribute("class");       
	msg.textContent= (stim_type === '1') ? 'The number of stimuli (images) in the task' : 'The number of stimuli (random alphanumeric characters) in the task';   
}  


// set Type of stimulus masking depending on Stimulus type
function setMask() 
{
	document.getElementById('stripe_cnt').disabled= false;
	var e= document.getElementById('mtype');
	var ind= e.selectedIndex;
	if (document.getElementById('stype').value === "0")  
	{
		for (var i= e.options.length-1; i>3; i--)   {     
       e.options[i].disabled= true;
    }         
		if (ind > 3) {
			e.selectedIndex= 0;
		}	
	}
	else  
	{
		for (var i=0; i< e.options.length; i++)   {
			e.options[i].disabled= false; 
		}
	}
}	


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById("is_valid").textContent=  (dirty===true) ? "All your settings have been successfully updated" : "";  
    document.getElementById("is_valid").setAttribute("style", "text-align:center;color:#000000"); 
  }
  else  {
    document.getElementById("is_valid").textContent="Some of your options are incorrectly set. They were not updated. These settings are marked in red."; 
    document.getElementById("is_valid").setAttribute("style", "text-align:center;color:#ff0000;font-weight:bold"); 
  } 	
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// check for validity.valid object in all input type="number" + other internal app checks.see: setPar()
function optionsValid()  
{     
  var res= document.getElementById('item_cnt').validity.valid && document.getElementById('stripe_cnt').validity.valid && document.getElementById('slevel').validity.valid;
	if (document.getElementById('stype').value == '0' && document.getElementById('alfa_cnt').validity.valid == false)	{ 
		res= false; 
	}	
	if (stype === "1" && document.getElementById('preview').getElementsByTagName('img').length === 0) {	
	  res= false;
	}
  return res;               
}


// start Task (research) 
function startTask(e)   
{     
  if (stype === "1" && document.getElementById('preview').getElementsByTagName('img').length === 0) {		 
		document.getElementById("is_valid").textContent= 'You have not uploaded images to use for stimuli. The number of images is zero. The requested action has been aborted.'; 
    document.getElementById("is_valid").setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold');
    return;		
	}  
	
	setPar();
 
	setStim();
	
	var canvas= document.getElementById("canvas");
	document.getElementById("result_wrapper").style.display="none";
  document.getElementById("end").style.display= "none"; 
  document.getElementById("start").style.display= "none";              
  document.getElementById("footer").style.display= "block"  	
	document.getElementById("progress").style.display= "block"; 
  document.getElementById("progress").setAttribute("style", "width:100%");
	document.getElementById('is_correct').textContent= '';
	if (feedback)  {
		document.getElementById("is_correct").style.display= 'block';
		document.getElementById("items_left").style.display= 'block';
	}
	else  {
		document.getElementById("is_correct").style.display= 'none';
		document.getElementById("items_left").style.display= 'none';
	}
	canvas.style.display= 'block'; 
	canvas.style.backgroundColor= '#ffffff';
	var ctx = canvas.getContext('2d');
	ctx.strokeStyle= "#000000";
	if (stype === '1' && validColor(mask_clr) ) 	{
		try  {
			canvas.style.backgroundColor= mask_clr;
		}
		catch(e) {			
		}				
	}	  
	if (stype === "0")
	{
		switch (alfa_type)
		{		
			case '1':  // Letters-lowercase 
				document.getElementById('answer').setAttribute('style','text-transform:lowercase');			
				break;
			case '2':  // Letters-UPPERCASE 
				document.getElementById('answer').setAttribute('style','text-transform:uppercase');			
				break;
			default:  // Digits ("0") or All digits and letters ("3")
				document.getElementById('answer').removeAttribute('style');	
		} 
	}
	else  {
		document.getElementById('answer').removeAttribute('style');	
		//document.getElementById('answer').setAttribute('style','text-transform:lowercase');		
	}
	
	is_task= true;
	
	setPage();      
		
	drawStim(true);     // draw (masked) first stimulus
      
}


// set internal application research parameters
function setPar()  
{  
  stim = [];   
	ans = [];  //  user answers for each item (dependent var)           
  percent = [];
  
  for (var j= 0; j< item_cnt; j++)   {
    ans[j]= '';
    percent[j]= 0;
  }
	item=0;
  clevel= slevel;  // current masking level == start level
	if (stype === "1" && item_cnt > document.getElementById('preview').getElementsByTagName('img').length) {		 
		item_cnt= document.getElementById('preview').getElementsByTagName('img').length;
		document.getElementById("is_valid").textContent= "The number of items was greater than the number of images. The number of items is automatically reduced to: " + item_cnt; 
    document.getElementById("is_valid").setAttribute("style", "text-align:center;color:#ff0000;font-weight:bold"); 
	}  
	if (stype === "0" && parseInt(mtype,10) > 3)
	{
		mtype= '0';
		document.getElementById("is_valid").textContent="The selected masking method is not available for the stimulus type: Alphanumeric characters. Automatically set to default value."; 
    document.getElementById("is_valid").setAttribute("style", "text-align:center;color:#ff0000;font-weight:bold"); 
	}		
}


// generates task stimuli 
function setStim()	
{
	if (stype === '0')  // alphanumeric characters are selected
	{  
		var s, previous;		
		for (var j= 0; j<item_cnt; j++)   
		{		
			do
			{	
			  s='';
				switch (alfa_type)
				{
					case '0':  // digits 		
						for (var i=0; i<alfa_cnt; i++)  {
							s+= getRndInt(0,9).toString();
						}			  
						break;			
					case '1':  // Letters-lowercase 
						s= lowerStrGen(alfa_cnt);			
						break;
					case '2':  // Letters-UPPERCASE 
						s= upperStrGen(alfa_cnt);			
						break;
					default:  // All digits and letters 
						s= allStrGen(alfa_cnt);
				} 
			} while (s === previous);
			stim[j]=s;
			previous=s;		
		}
	}
	else  // images are selected
	{
		var image;
		for (var j= 0; j<item_cnt; j++)   {
			image= document.getElementById('preview').getElementsByTagName('img')[j];
			stim[j]= image.title ? image.title : 'image ' + parseInt(j+1);  // images file names array => document.querySelector('input[type=file]').files[i].name
			//var is_title= image.getAttribute('data-title');
			//stim[j]= is_title ? is_title : 'item ' + parseInt(j+1);  // images file names array => document.querySelector('input[type=file]').files[i].name
		}					
	}
}


// generates a random string from lowercase letters
function lowerStrGen(cnt)
{ 
  var all = "abcdefghijklmnopqrstuvwxyz";
  var txt = "", tmp= all.length;
  for (var i = 0; i < cnt; i++)  {
    txt += all.charAt(Math.floor(Math.random() * tmp));
	}
  return txt;
}


// generates a random string from uppercase letters 
function upperStrGen(cnt)
{ 
  var all = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  var txt = "", tmp= all.length;
  for (var i = 0; i < cnt; i++)  {
    txt += all.charAt(Math.floor(Math.random() * tmp));
	}
  return txt;
}


// generates a random string from all (uppercase and lowercase) letters and digits
function allStrGen(cnt)
{ 
  var all = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
  var txt = "", tmp= all.length;
  for (var i = 0; i < cnt; i++)  {
    txt += all.charAt(Math.floor(Math.random() * tmp));
	}
  return txt;
}


// set canvas.height, canvas.width, global var font_size
// call in startTask() + each body onresize() when is_task == true => call this function
function setPage()  
{			
	// The read-only innerHeight property of the Window interface returns the interior height of the window in pixels, including the height of the horizontal scroll bar, if present.
  // window.innerHeight = The value of innerHeight is taken from the height of the window's layout viewport.  
	
	var h= window.innerHeight - document.getElementById("footer").offsetHeight; // window height - footer height
  document.getElementById("canvas").style.height= h + "px"; // The height property has effect only on block-level elements or on elements with absolute or fixed position.	
  canvas.height= h;
	canvas.width= window.innerWidth; 
	
	// set global window var => canvas font size (px) for drawText() function
  font_size= Math.round(Math.min(Math.floor(canvas.width/alfa_cnt), 0.5 * canvas.height));		
	
	drawStim(false);
}	


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
   
	if (document.getElementById('item_cnt').validity.valid)  {  
    if ( stype === '0' || (stype === '1' && parseInt(document.getElementById('item_cnt').value, 10) <= document.getElementById('preview').getElementsByTagName('img').length) ) {
			item_cnt= parseInt(document.getElementById('item_cnt').value, 10);     
		}
  } 
	 	 
  stype= document.getElementById('stype').value;     
	
  alfa_type= document.getElementById('alfa_type').value;
	
	if (document.getElementById('alfa_cnt').validity.valid) {
		alfa_cnt= parseInt(document.getElementById('alfa_cnt').value, 10);
  }		
	
	font_family= document.getElementById('font_family').value;
	
	mtype = document.getElementById('mtype').value;   
	
  if (document.getElementById('stripe_cnt').validity.valid)  {  
    stripe_cnt= parseInt(document.getElementById('stripe_cnt').value, 10);     
  } 
  
  if (document.getElementById('slevel').validity.valid)  {   
    slevel= parseInt(document.getElementById('slevel').value, 10);   
	}
	
	demo= document.getElementById('demo').checked ? true: false; 
	
	feedback= document.getElementById('feedback').checked ? true: false; 
	
	if (validColor(document.getElementById('mask_clr').value))  {		
		mask_clr= document.getElementById('mask_clr').value;
  }
		     
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}


// get random integer from min to max range
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1)) + min;
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// go to the next task if the ENTER key is pressed. see function init()
function answerKeyUp(e)  
{
	if (e.key === 'Enter' || e.keyCode === 13) {  
		next();
	}
}


// goto next item (stimulus)
function next()  
{  
  ans[item]= document.getElementById('answer').value;	
  // The \s meta character in JavaScript regular expressions matches any whitespace character: spaces, tabs, newlines and Unicode spaces. And the g flag tells JavaScript to replace it multiple times. 
	ans[item]= ans[item].replace(/\s/g, '');	 // const newStr = str.replace(regexp|substr, newSubstr|function)
	if (stype === '0')	{
		if (alfa_type === '1')  {
			ans[item]= ans[item].toLowerCase(); // "1" = Letters-lowercase;
		}	
		else if (alfa_type === '2')  {
			ans[item]= ans[item].toUpperCase(); //  "2" =Letters-UPPERCASE;
		}	  // alfa_type => "3" is case sensitive (all letters and digits); "0" => Digits
	}
	
	if ( (stype === '0' && ans[item] == stim[item]) || (stype === '1' && ans[item].toLowerCase() == 'y') )	{
		var is_cor= true; // answer is correct 
	}
	else  {
		var is_cor= false; // answer is not correct  
	}  	
	
	if (is_cor || clevel <= 0)	{  // goto next stimulus
		clevel= slevel;
		item++;
	}	
	else  {
		clevel--;
	}	
	if (document.getElementById('feedback').checked)  {
	  document.getElementById('is_correct').textContent= (clevel == slevel) ? 'A new stimulus is displayed.' : 'The same stimulus is displayed but less masked.';
		document.getElementById('is_correct').textContent+= is_cor ? ' The last answer was correct.' : ' The last answer was incorrect.';
	}	 
	if (item < item_cnt)  {         		 
    drawStim(true);     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


/* draw (repaint) stimulus
   upd_ans == true => update current answer.
	 upd_ans => escape update user answer => calls from body onresize="pageResize()" => setPage() */     
function drawStim(upd_ans) 
{  
  if (upd_ans)  {  
		document.getElementById('answer').value=  (demo  && stype === "0") ?  stim[item] : '';  // show correct answer or delete previous answer
		document.getElementById('answer').focus();
	}
	
	clear();  // clear canvas
		          
	if (stype === '0')
	{     
		drawText(stim[item]); 		
					
		switch (mtype)
		{  // mytpe == Type of stimulus masking
			case '0':
				maskHStripeBlack(per_arr[clevel], stripe_cnt);  
				break;
			case '1':
				maskVStripeBlack(per_arr[clevel], stripe_cnt);  
				break;
			case '2':
				maskHorBlack(per_arr[clevel]);  
				break;				
			case '3':
				maskVerBlack(per_arr[clevel]); 
				break;		
		}			
	}
	else	
	{		
    var img = document.getElementById('img_cont');  // img_cont.style="display:none"		
				//img.crossOrigin= 'anonymous'; // If crossOrigin is an empty string (""), the anonymous mode is selected.
		img.src= document.getElementById('preview').getElementsByTagName('img')[item].src;   		
	}		
 	
	updateProgress();	
		  
}


// draw text (stimulus) on canvas 
function drawText(txt)  
{
	var canvas= document.getElementById('canvas');
  var ctx= canvas.getContext('2d'); 
	//ctx.fillStyle = "#ffffff"; 
	//ctx.clearRect(0, 0, canvas.width, canvas.height);      // clear canvas
	var x= canvas.width/2;
	var y= canvas.height/2;	     
	ctx.fillStyle = "#000000";  // line color 	
	var dx= Math.floor(0.5 * canvas.width);  
	var dy= Math.floor(0.5 * canvas.height); 
	var rad= Math.floor(Math.min(dx,dy)/2);  // text radius 	                                                                   
  /* https://developer.mozilla.org/en-US/docs/Web/CSS/font-family   
	   https://www.w3schools.com/cssref/css_websafe_fonts.asp	 
		 examples: ctx.font = "72px monospace, Courier";  	ctx.font = "5em Georgia";  
	*/
	ctx.font= font_size.toString() + 'px ' + font_family;  
	ctx.textBaseline= "bottom";  ctx.textBaseline = "middle";    
	ctx.textAlign= "center";
	
	ctx.fillText(txt, x, y); 	
}	


/* draw image from ../images directory into canvas object (html container)
   see:  https://coderwall.com/p/jzdmdq/loading-image-from-local-file-into-javascript-and-getting-pixel-data
	 https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/drawImage
	 https://stackoverflow.com/questions/17049176/ns-error-not-available-component-is-not-available?rq=1  */
function loadImage()  
{	
	var canvas= document.getElementById('canvas'), ctx= canvas.getContext('2d');
	var max_w= Math.round(0.95*canvas.width), max_h= Math.round(0.95*canvas.height);
	var img= document.getElementById('img_cont');  // image stimulus html container
	var x0, y0, w, h;
	if (img.naturalWidth <= max_w && img.naturalHeight <= max_h)  
	{ 
		x0= Math.floor(canvas.width/2 - img.naturalWidth/2);
		y0= Math.floor(canvas.height/2 - img.naturalHeight/2);
		w= img.naturalWidth;
		h= img.naturalHeight;					
	}
	else  
	{  	// rescale image			
		w= img.naturalWidth/2;  // decrease width 100%
		h= img.naturalHeight/2; // decrease height 100%
		while (w > max_w || h > max_h)  {
			w= w/2;
			h= h/2;
		}  				
		x0= Math.floor(canvas.width/2 - w/2);
		y0= Math.floor(canvas.height/2 - h/2);						
	}	
	img.width=w;  
	img.height=h;				
	bounds ={left:x0, top: y0, width: w, height: h};	

	if (!is_task) {
	  return;	
	}	
  //clear();  // clear canvas
	
	ctx.drawImage(img, x0, y0, w, h);	
	
	if (clevel < 0)  {
		return;
	}
	
	switch (mtype)
	{  // mytpe == Type of stimulus masking
		case '0':
			maskHStripe(per_arr[clevel], stripe_cnt);  // call function maskHStripeBlack(per, step)
			break;
		case '1':
			maskVStripe(per_arr[clevel], stripe_cnt);  // call function maskVStripeBlack(per, step)
			break;
		case '2':
			maskHor(per_arr[clevel]);  // function maskHor(per)
			break;				
		case '3':
			maskVer(per_arr[clevel]); // function maskVer(per)
			break;	
		case '4':
			maskGrid(per_arr[clevel], stripe_cnt); // Creates a grid effect => function maskGrid(per, cnt)
			break;		
		case '5':
			maskRect(per_arr[clevel], stripe_cnt); // Rectangles or inverse grid effect => function maskRect(per, cnt)
			break;	
		case '6':
			maskCircle(per_arr[clevel], stripe_cnt); // circles => function maskCircle(per, cnt)
			break;			
	}	
}


// Update Progress bar on footer
function updateProgress()
{
  var percent = 100*(item_cnt-item)/item_cnt;  // Math.round();      
	document.getElementById("progress").setAttribute("style", "width:"+percent+"%");    // !important
	document.getElementById("items_left").textContent= feedback ? "Item: " + (item + 1) + " (Masking level:" + clevel + ").  Remaining items: " + (item_cnt - item-1) : '';
  if (item === item_cnt-1)	{		
		document.getElementById("next").style.backgroundImage = "url('img/end.png')";
		document.getElementById("next").title = 'End of task';
	}
	else	{
		document.getElementById("next").style.backgroundImage = "url('img/forward.png')"; 
		document.getElementById("next").title = 'Next task';
	}
}	


// end of task (research)
function theEnd()  
{   
  is_task= false;  
 	clear();
	document.getElementById("canvas").style.display="none";
	document.getElementById("footer").style.display="none";
  document.getElementById("end").style.display="block";   
  
	// compute total number of recognized stimulus (correct) 
  var cor = 0, tmp = [], tmp_v = [], tmp_i = [];   
	for (var i=0; i<item_cnt; i++)  
	{
		if ( (stype === '0' && ans[i] == stim[i]) || (stype === '1' && ans[i].toLowerCase() == 'y') )	
		{
			cor++;  // number of recognized stimulus (correct answers)
			//ans[i]= (stype === '1' && ans[i] === 'y') ? 'Correct' : 'Incorrect';  // recode answer 'y' for stimulus type = images
			tmp[i] = [ parseInt(i+1,10), stim[i], ans[i], 'Yes', 100-percent[i], percent[i] ];  // tmp[i] => ['Item', 'Stimulus' , 'Answer', 'Correct=Yes', 'Visible pixels %', 'Invisible pixels %']); 
			tmp_v.push(100-percent[i]);  		  // Percentage of visible (non-mask) pixels 
			tmp_i.push(percent[i]);   // Percentage of invisible (mask) pixels			
		}
		else  {  // unrecognized stimulus
			tmp[i]= [ parseInt(i+1,10), stim[i], ans[i], 'No',  '', '' ];  // tmp[i] => ['Item', 'Stimulus' , 'Answer', 'Correct=No', '', '']); 
		}			  		
	}		
				
  document.getElementById("result_wrapper").style.display="block";  
	createTableArray(document.getElementById('table1'), tmp,  ['Item', 'Stimulus' , 'Answer', 'Correct', 'Visible pixels %', 'Invisible pixels %']);  
	document.getElementById("table1").style.display="block";  	
	
	// calculate decriptive statistics for Visible and Invisible % pixels => average and sample standard deviation..	
	var m_v =  (cor > 0) ? mean(tmp_v).toFixed(3) :  'Cannot be calculated';  // mean - Percentage of Visible pixels
	var sd_v = (cor > 1) ? stDev(tmp_v, m_v).toFixed(3) : 'Cannot be calculated'; // standard deviation Percentage of Visible pixels

  var m_i = (cor > 0) ? mean(tmp_i).toFixed(3) :  'Cannot be calculated';  // mean - Percentage of Invisible pixels
	var sd_i = (cor > 1) ? stDev(tmp_i, m_i).toFixed(3) : 'Cannot be calculated'; // standard deviation - Percentage of Invisible pixels
  
	tmp=[];
	tmp[0]= ['Visible pixels %', m_v, sd_v, cor];
	tmp[1]= ['Invisible pixels %', m_i, sd_i, cor];			
	createTableArray(document.getElementById('table2'), tmp, ['Variable', 'Average', 'Sample Standard Deviation', 'N']);   
	
	// draw - repaint "chart" => mean for invisible pixels (m_i) and visible pixels (m_v)	
	if (cor > 0)  {
		document.getElementById('chart').style.display='block';
		document.getElementById('chart').title= "Red = Visible pixels (%) and Green = Invisible (masked) pixels (%)";
		document.getElementById('chart_title').textContent= 'Arithmetic mean for Visible pixels % (red) = ' + m_v + ' Arithmetic mean for Invisible (masked) pixels % (green) = ' + m_i;			
		document.getElementById("chart_vis").setAttribute("style", "width:"+m_v+"%");
		document.getElementById("chart_inv").setAttribute("style", "width:"+m_i+"%");    // !important
	}
	else  {  // chart is not visible for cor == 0 
		document.getElementById('chart_title').textContent= 'Chart is disabled (no correct answers)';
		document.getElementById('chart').style.display='none';
		document.getElementById('chart').title="";
	}
	
	try 
  {
    if (store) {
			if ( isNaN(m_i) ) { m_i= '';  }   // recode 'Cannot be calculated' => '' for browser storage
      if ( isNaN(sd_i) ) { sd_i= '';  } // recode 'Cannot be calculated' => '' for browser storage
			setStorage(cor, m_i, sd_i);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }    	

}


// return arithmetic mean (mean, average) of array (arr) 
function mean(arr)  {	
	var sum=0;
	for (var i in arr)  {
		sum+= parseFloat(arr[i]);
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation
// arr = array argument; mean = average (mean)
function stDev(arr, mean) {
	var len= arr.length, sum= 0, res;
	for (var i=0; i<len; i++) {
	  sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	var today = new Date();
	var dd = today.getDate();
	var mm = today.getMonth()+1; 
	var yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* set user data into window.localStorage
	 cor  => number of recognized stimulus (correct answers)
	 m_i_ => arithmetic mean for Invisible (masked) pixels
	 sd_i => sample Standard Deviation for Invisible (masked) pixels  */
function setStorage(cor, m_i, sd_i)   
{  
  var oldItems = JSON.parse(localStorage.getItem("redundancy-web-v1.0")) || []; 
  var id= oldItems.length+1;   
  var cnt= (document.getElementById('mtype').value === '2' || document.getElementById('mtype').value === '3')  ? '' : stripe_cnt;			
	// set new "record"
  var newItem = {
      'id': id,
      'date': setDate(),   //new Date(),  
			'items': item_cnt,     
      'correct': cor,                   
      'M-invisible': m_i, 
			'SD-invisible': sd_i,  			      
			'stimulus-type' : stype,
			'alfa-type': stype === '0' ? alfa_type : '', 
			'alfa-cnt':  stype === '0' ? alfa_cnt  : '', 
			'mask-type': mtype,
			'raster' : cnt,
			'start-level': slevel,
			'feedback': (document.getElementById('feedback').checked) ? '1' : '0',
      'demo' : (demo && stype != "1") ? "1": "0"        
  };

  oldItems.push(newItem);

  localStorage.setItem("redundancy-web-v1.0", JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
  	
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  var col = [];
  for (var i = 0; i < my_json.length; i++) {
    for (var key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  var table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  var tr = table.insertRow(-1);                   // TABLE ROW.
  for (var i = 0; i < col.length; i++) {
      var th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (var i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (var j = 0; j < col.length; j++) {
      var tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	var table= document.createElement("table");  // html table
  var tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (var i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (var i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (var j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (var i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear redundancy-web web browser storage data
function clrStorage()  
{  
  var cnt= localStorage.getItem("redundancy-web-v1.0")  ?  JSON.parse(localStorage.getItem("redundancy-web-v1.0")).length : 0;                                                        
  if (cnt === 0)  {
    alert("You have no Redundancy-web data to delete");
    return;
  }  
  if ( !window.confirm("You have a total of " + cnt  + " Redundancy-web results stored in your web browser. Continue?") ) { 
   return;
 }                                             
 if (window.confirm("Delete total saved Redundancy-web results:  "  + cnt + ". This action cannot be undone. Final confirmation?") ) 
 { 
   window.localStorage.removeItem("redundancy-web-v1.0"); 
   document.getElementById("store_cnt").textContent= "All data has been deleted. No data";
 }  
}


// clear the canvas
function clear()  {
	var canvas = document.getElementById('canvas');
	var ctx = canvas.getContext('2d');	 
	ctx.fillStyle = "#ffffff";
	ctx.clearRect(0, 0, canvas.width, canvas.height);
	//ctx.fill();   
}	


/* 
		mask black canvas pixels(set A-value transparency) with horizontal stripes (lines)
		per = float between 0..1. "Percent" of black pixels to change (set A-value transparency)
    per = 1 => mask all black pixels, 0 => mask none, 0.5 => mask 50% black pixels 
		step = number of steps (raster). Number of image segments to be masked. minimum step = 1 
*/
function maskHStripeBlack(per, step)
{	
	var canvas= document.getElementById('canvas'), ctx= canvas.getContext('2d'), imageData= ctx.getImageData(0, 0, canvas.width, canvas.height), pixels= imageData.data, len= pixels.length; 	
	var rows= analyzeRows(canvas);  //  rows = number of black pixels in each row, last element in array is total count of black pixels	
	var maxim= Math.round(per * rows[rows.length-1]);  // total number of black pixel to change
	var k = Math.round(maxim/step);   // total number of pixel to change in each step 
	var j = Math.round((rows[rows.length-1]-maxim)/(step+1));  // number of black pixel to ignore between each step
  var	total= 0, ignore_cnt= 0, changed_cnt= 0, total_changed=0;	// counters	
	for (var i= 0; i< len; i+= 4) 
	{
		if ( pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 )
		{  // this is target color (black)
			total++;  // total number of black pixel in canvas (target color)
		  if (ignore_cnt <= j) 
			{
				ignore_cnt++;
				continue;
			}
			else if (changed_cnt < k)  
			{
				pixels[i+3]= 0;   // mask this pixel => set A value (transparency)			
				changed_cnt++;	  // number of changed pixels in this step
				if (changed_cnt == k)  {  // reset counters for next step
					total_changed+= changed_cnt;  // total number of changed (mask) black pixels
					changed_cnt= 0;   
					ignore_cnt= 0;
				}
			}				
		}								
	}			
	ctx.putImageData(imageData, 0, 0);  // Draw image data to the canvas
	percent[item]= (100*total_changed/total).toFixed(3);	
}	


/* 
		mask black canvas pixels(set A-value transparency) with vertical stripes (lines)
		per = float between 0..1. "Percent" of black pixels to change (set A-value transparency)
    per == 1 => mask all black pixels, 0 => mask none, 0.5 => mask 50% black pixels 
		step = number of steps (raster). Number of image segments to be masked. minimum step = 1 
*/
function maskVStripeBlack(per, step)
{	
	var canvas= document.getElementById('canvas'), ctx= canvas.getContext('2d'), w= canvas.width, h= canvas.height, imageData= ctx.getImageData(0, 0, w, h), pixels= imageData.data; 	
	var black= getBlackCount(canvas);  // total number of black pixels in canvas object
	var maxim= Math.round(per * black);  // total number of black pixel to change
	var k = Math.round(maxim/step);   // total number of pixel to change in each step 
	var j = Math.round((black-maxim)/(step+1));  // number of black pixel to ignore between each step	
  var	total= 0, ignore_cnt= 0, changed_cnt= 0, total_changed=0, ind;	// counters
	// Iterate through every pixel by columns (vertical direction)
	for (var col = 0; col < w; col++) 
	{	
		for (var row = 0; row < h; row++) 
		{		
			ind= getColorIndex(col, row, w);	
			if ( pixels[ind] == 0 && pixels[ind+1] == 0 && pixels[ind+2] == 0 && pixels[ind+3] > 0 )
			{	 // this is target color (black)
				total++;  // total number of black pixel in canvas (target color)
				if (ignore_cnt <= j) 
				{
					ignore_cnt++;
					continue;
				}
				else if (changed_cnt < k)  
				{
					pixels[ind + 3]= 0;   // mask this pixel => set A value (transparency)			
					changed_cnt++;	  // number of changed pixels in this step
					if (changed_cnt == k)  
					{  // reset counters for next step
						total_changed+= changed_cnt; // total number of changed (mask) black pixels
						changed_cnt= 0;   
						ignore_cnt= 0;
					}
				}				
			}								
		}
	}		
	ctx.putImageData(imageData, 0, 0);	 // draw image data to the canvas
	percent[item]= (100*total_changed/total).toFixed(3);	
}	


/*	Masks (sets invisible) black pixels inside the image. Cut one piece - horizontally
    The image is always masked in one piece - horizontally
		per = Percentage of black pixels to be set to invisible  
		per = float between 0..1. "Percent" of black pixels to change (set A-value transparency) */		
function maskHorBlack(per)
{
	var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d'), imageData = ctx.getImageData(0, 0, canvas.width, canvas.height), pixels = imageData.data, len= pixels.length; 
	var total= getBlackCount(canvas), k= 0;   // total number of black pixels, changed pixel counter  
	var maxim= Math.round(per * total);  // maximum allowed total number of black pixel to change 
	for (var i= 0; i< len; i+= 4) 
	{
		if ( pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 && k <= maxim)
		{  // this is target color (black)
			pixels[i + 3]= 0;  // A value		
			k++;   // changed black pixels pixel counter							 			
		}
	}	
	ctx.putImageData(imageData, 0, 0);	
	percent[item]= (100*k/total).toFixed(3);		
}	


/*	Masks (sets invisible) black pixels inside the image. Cut one piece - vertically
    The image is always masked in one piece - vertically
		per = Percentage of black pixels to be set to invisible 
		per = float between 0..1. "Percent" of black pixels to change (set A-value transparency) */
function maskVerBlack(per)
{
	var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d'), w= canvas.width, h= canvas.height;	  
	var imageData = ctx.getImageData(0, 0, w, h), pixels = imageData.data;
	var total= getBlackCount(canvas);   // total number of black pixels 
	var maxim= Math.round(per * total);  // maximum allowed total number of black pixel to change
	var k= 0, ind;  // counters. k == changed pixel counter
	// Iterate through every pixel by columns (vertical direction)
	for (var col= 0; col< w; col++) 
	{		
		for (var row= 0; row< h; row++) 
		{			
			if (k < maxim)  
			{
			  ind= getColorIndex(col, row, w);	
				if ( pixels[ind] == 0 && pixels[ind+1] == 0 && pixels[ind+2] == 0 && pixels[ind+3] > 0 )
				{ // this is target pixel color == black
					pixels[ind+3]= 0;  // Alpha value		
				  k++;   // changed black pixel counter
				}
			} 			
		}
	}	
	ctx.putImageData(imageData, 0, 0);	
	percent[item]= (100*k/total).toFixed(3);	
}	


// https://developer.mozilla.org/en-US/docs/Web/API/Canvas_API/Tutorial/Pixel_manipulation_with_canvas
// x = col, y = row, width = canvas.width
function getColorIndex(x, y, width) 
{
  var ind = y * (width * 4) + x * 4;
  return ind;
}


/* use only for black pixels search
   e == canvas object
   return res = number of black pixels in each row, last element in array is total count of black pixels  */
function analyzeRows(e)
{
	var ctx = e.getContext('2d'), w = e.width, pixels= ctx.getImageData(0, 0, w, e.height).data, len= pixels.length;
	var cnt= 0, row= 0, j= 0, black= 0, res=[];  // res => number of pixels in each row  
	for (var i = 0; i != len; i += 4) 
	{		
		if ( pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			j++;      			
		}   	
	  cnt++;    // total pixel counter
		if (cnt % w == 0) {   // enter new row
			res[row]= j;			
			row++;
			black+=j;
			j= 0;			
		}	
	}
	res.push(black);  // last element in array => total number of black pixels
	return res;
}	

 
/* 	e == canvas object
		return total number of black pixels in canvas object */
function getBlackCount(e)
{
	var ctx = e.getContext('2d'), pixels= ctx.getImageData(0, 0, e.width, e.height).data, len= pixels.length, cnt= 0;  	
	for (var i= 0; i!= len; i+= 4) 
	{		
		if ( pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {			
			cnt++;    // total black (target) color pixel counter  			
		}   		  		
	}
	return cnt; // cnt => total number of black pixels in canvas object
}	


/*  Clear image segments horizontally - set horizontal stripes (lines) in image 
    Use this function only for image (user custom) stimulus 
		per = float between 0..1 => Percentage of image area to be clear (set to invisible)
		cnt = number of stripes. Number of image segments to be clear. minimum cnt = 1  */
function maskHStripe(per, cnt)
{	
  var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d');
	var total= bounds.width * bounds.height;   // = img.width * img.height => total area (img= document.getElementById('img_cont');	
	var area= per * total / cnt;  // area to clear => for each stripe (only one stripe)
	var h= Math.round(area/bounds.width);  // height for each stripe =>  h * bounds.width = area => h = area/bounds.width;	      
	var dy = Math.round( (bounds.height - cnt * h) / (cnt+1) );	 // vertical spacing between stripes 	
	for (var i=0; i<cnt; i++) {
		ctx.clearRect(bounds.left, Math.round(bounds.top  + i*h + (i+1)*dy), bounds.width, h);		
	}			
	percent[item]= (100*cnt*h*bounds.width/total).toFixed(3);  // percent[item]= (100*cnt*area/total).toFixed(3);	
}	


/*  Clear image segments vertically - set vertical stripes (lines) in image 
    Use this function only for image stimulus
		per = float between 0..1 => Percentage of image area to be clear (set to invisible)
		cnt = number of stripes. Number of image segments to be clear. minimum cnt = 1  */
function maskVStripe(per, cnt)
{	
  var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d');
	var total= bounds.width * bounds.height;   // = img.width * img.height => total area (img= document.getElementById('img_cont');	
	var area= per * total / cnt;  // area to clear => for each stripe (only one stripe)	
	var w= Math.round(area/bounds.height);  // width for each stripe =>  w * bounds.height = area => w = area/bounds.height;	  
  var del_x = Math.round( (bounds.width  - cnt * w) / (cnt+1) );  // horizontal spacing between stripes
	for (var i=0; i<cnt; i++) 
	{  // i=> y-axis == column
		var x = Math.round(bounds.left + i*w + (i+1)*del_x);
		ctx.clearRect(x, bounds.top, w, bounds.height);   // context.clearRect(x,y,width,height); 
	}	
	percent[item]= (100*cnt*w*bounds.height/total).toFixed(3);  // percent[item]= (100*cnt*area/total).toFixed(3);	
}	


/*	Clear (sets invisible) pixels inside the image. Cut in one piece - horizontally.   
    Use this function only for image stimulus
    The image is masked (with clearRect) in one piece - horizontally
		per = float between 0..1 => Percentage of image area to be clear or set to invisible */
function maskHor(per)
{
	var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d');
	var total= bounds.width * bounds.height;   // = img.width * img.height => total area (img= document.getElementById('img_cont');	
	var area= per * total;  // area to clear
	var h= Math.round(area/bounds.width);  // stripe height =>  h * bounds.width = area => h = area/bounds.width;	  
	var ys= bounds.top + bounds.height/2;  // y-center	 //var xs= bounds.left + bounds.width/2;  // x-center 
	ctx.clearRect(bounds.left, ys-h/2, bounds.width, h);
	percent[item]= (100*h*bounds.width/total).toFixed(3);  // percent[item]= (100*area/total).toFixed(3);		
}	

/*	Clear (sets invisible) pixels inside the image. Cut in one piece - vertically.
    Use this function only for image stimulus   
    The image is masked (with clearRect) in one piece - vertically
		per = Percentage of image area (pixels) to be set to invisible 
		per = float between 0..1. "Percent" of pixels to clear */
function maskVer(per)
{
	var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d');
	var total= bounds.width * bounds.height;   // = img.width * img.height => total area (img= document.getElementById('img_cont');	
	var area= per * total;  // area to clear
	var w= Math.round(area/bounds.height);  // stripe width =>  w * bounds.height = area => w = area/bounds.height;
	var xs= bounds.left + bounds.width/2;  // x-center   //var ys= bounds.top + bounds.height/2;  // y-center	 
	ctx.clearRect(xs-w/2, bounds.top, w, bounds.height);	
	percent[item]= (100*w*bounds.height/total).toFixed(3);  // percent[item]= (100*area/total).toFixed(3);
}	


/*  Clear image segments horizontally + vertically - Creates a grid effect 
		Use this function only for image stimulus
		per = float between 0..1 => Percentage of image area to be clear (set to invisible)
		cnt = number of stripes (raster). Number of image segments to be clear. minimum cnt = 1  */
function maskGrid(per, cnt)  
{
  var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d');
	var total= bounds.width * bounds.height;   // = img.width * img.height => total area
	var area= per * total / (2*cnt);  // area to clear => for each stripe (only one stripe)
	var h= Math.round(area/bounds.width);  // height for each horizontal stripe =>  h * bounds.width = area => h = area/bounds.width;	  
	var w= Math.round(area/bounds.height);  // width for each vertical stripe =>  w * bounds.height = area => w = area/bounds.height;	
	var dy= bounds.height/(cnt+1), y0=bounds.top;    // dy = vertical increament (y-step in canvas coord)     
	var dx= bounds.width/(cnt+1), x0= bounds.left;   // dx = horizontal increament (x-step in canvas coord)
	// set (delete) horizontal stripes
	for (var i= 1; i<= cnt; i++) {
		ctx.clearRect(bounds.left, Math.round(y0+i*dy-h/2), bounds.width, h);
	}		
	// set (delete) vertical stripes
	for (var i= 1; i<= cnt; i++) {
		ctx.clearRect(Math.round(x0+i*dx-w/2), bounds.top, w, bounds.height);
	}
	var hor_area = cnt*h*bounds.width;  // total horizontal stripe area
	var ver_area = cnt*w*bounds.height; // total vertical stripes area
	var stripe_area = hor_area + ver_area - (cnt*cnt*w*h);  // Total stripes area => subtract cross section (for horizontal and vertical stripes)
	percent[item]= (100*stripe_area/total).toFixed(3); 		
}	

	
/* 	Masks image in the shape of a rectangle (inverse grid)
    Use this function only for image stimulus
		per = float between 0..1 => Percentage of image area to be clear (set to invisible)
		cnt = raster number. Number of image segments to be clear. minimum cnt = 1  */
function maskRect(per, cnt)  
{
	var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d');
	var w = Math.round(per*bounds.width/cnt);   // width for each rectangle 
	var h = Math.round(per*bounds.height/cnt);  // height for each rectangle
	var del_x = Math.round( (bounds.width  - cnt * w) / (cnt+1) );  // horizontal spacing between rectangles
  var del_y = Math.round( (bounds.height - cnt * h) / (cnt+1) );	 // vertical spacing between rectangles  	
	var x, y; // current coordinates (x,y)
	for (var i=0; i<cnt; i++) 
	{  // i=> y-axis == column
		for (var j=0; j<cnt; j++) 
		{  // j => x-axis == row
			x = Math.round(bounds.left + j*w + (j+1)*del_x);
			y = Math.round(bounds.top  + i*h + (i+1)*del_y);
			ctx.beginPath();
			ctx.clearRect(x, y, w, h);  // context.clearRect(x,y,width,height); 
			//ctx.stroke(); 
		}
	}	
	// Total masked area => cnt*cnt*w*h. Total image area => bounds.width * bounds.height
	var res = ( 100 * cnt * cnt * w * h / (bounds.width * bounds.height) ).toFixed(3); 
	percent[item]= res;	
}	


/* 	Masks image in the shape of a circles
    Use this function only for image stimulus
		per = float between 0..1 => Percentage of image area to be clear (set to invisible)
		cnt = raster number. Number of image segments to be clear. minimum cnt = 1  */
function maskCircle(per, cnt)  
{ 
	var canvas = document.getElementById('canvas'), ctx = canvas.getContext('2d');
	var w = Math.round(per*bounds.width/cnt);   // width for each image section (raster) 
	var h = Math.round(per*bounds.height/cnt);  // height for each image section (raster)
	var del_x = Math.round( (bounds.width  - cnt * w) / (cnt+1) );  // horizontal spacing between sections
  var del_y = Math.round( (bounds.height - cnt * h) / (cnt+1) );	 // vertical spacing between sections  	
	var x, y; // current coordinates (x,y)
	var r= Math.round( 0.5 * Math.min(w,h) );   // circle radius
	ctx.fillStyle= canvas.style.backgroundColor;  // ctx.fillStyle= mask_clr;
	for (var i=0; i<cnt; i++) 
	{  // i=> y-axis == column
		for (var j=0; j<cnt; j++) 
		{  // j => x-axis == row
			x = Math.round(bounds.left + j*w + (j+1)*del_x + w/2);  // circle x-center 
			y = Math.round(bounds.top  + i*h + (i+1)*del_y + h/2);  // circle y-center 
			ctx.beginPath(); 
			ctx.arc(x, y, r, 0, 2*Math.PI, false);  // ctx.arc(x,y,r,sAngle,eAngle,counterclockwise);           			       
			ctx.fill();  
		}
	}	
	// Total masked area => cnt*cnt*r*r*Math.PI. Total image area => bounds.width * bounds.height
	var res = ( 100 * (cnt * cnt * r * r * Math.PI) / (bounds.width * bounds.height) ).toFixed(3); 
	percent[item]= res;	
}


/* EventListener for input type="file" onchange. see html: document.getElementById("add_file")
   When the user selects image files
   https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsDataURL */
function createImages() 
{
  var preview = document.querySelector('#preview');
  var files = document.querySelector('input[type=file]').files;
  var cnt= document.getElementById('preview').getElementsByTagName('img').length;  
	function readAndPreview(file) 
	{    
    if ( /\.(jpe?g|png|gif|bmp)$/i.test(file.name) ) {  // Make sure `file.name` matches our extensions criteria
      var reader = new FileReader();
      reader.addEventListener("load", function () {
        var image = new Image();              
        image.src = this.result;        
				image.setAttribute('class', 'flex-child');   //image.height = 100;
				image.title = 'item ' + parseInt(cnt+1) + '=' + file.name;
				//image.setAttribute('data-title', 'item ' + parseInt(cnt+1)); 
				// drag and drop  
				image.id= 'img' + parseInt(cnt+1); 
				image.setAttribute("draggable", "true"); 
				image.setAttribute("ondragstart", "onDragStart(event)");   				
				image.setAttribute("ondragover", "onDragOver(event)");
				image.setAttribute("ondrop", "onDrop(event)");
				
				preview.appendChild(image);
				cnt++;
				if (cnt>0) document.getElementById('img_info').style.display= 'block';
				document.getElementById('item_cnt').value= cnt; 
				itemsValid();   //document.getElementById('item_cnt_msg').textContent= ''; 
      }, false);
      reader.readAsDataURL(file); 
    }
  }
  if (files) {
    [].forEach.call(files, readAndPreview);
  }
 
	//document.getElementById('item_cnt').value= files.length;  	
}


// ondragstart event for stimulus type= images
function onDragStart(ev)  {   
  ev.dataTransfer.setData('text', ev.target.id);
}


// ondragover() event for stimulus type= images
function onDragOver(ev)  {  
  ev.preventDefault();
}


// ondrop() event for stimulus type= images
function onDrop(ev)  
{  
	ev.preventDefault(); 
  var my_id = ev.dataTransfer.getData('text');    // node1.id
  var node1= document.getElementById(my_id);    // node1= DRAG node pointer           
	var node2= ev.target;   //  node2= DROP node pointer: above  <ol> or <li> ??  => [object HTMLOListElement] or [object HTMLLIElement] => this is drop node      
	if ( node2.tagName === "IMG" )  {      // drop mouse pointer above <li> 
		node2.parentNode.insertBefore(node1, node2);   // move (insert) node1 before node2
	}	
  toogleInfo();
}


// remove image file from container (stimulus type: images). see html source
function delItem(ev)
{ 
	ev.preventDefault(); 
  var my_id= ev.dataTransfer.getData('text');    // node1.id
  var node1= document.getElementById(my_id);    // node1= DRAG node pointer             		
	node1.parentNode.removeChild(node1);   // delete ex node2
	toogleInfo();	
}


// auto set item_cnt in task based on current number of images stimulus
// set visibility for small info div: "Drop image in this line here to delete" and "Use drag and drop to move files (stimulus)"
function toogleInfo()  
{
	var img_cnt= document.getElementById('preview').getElementsByTagName('img').length;  // current number of images stimulus
	document.getElementById('item_cnt').value= img_cnt;
  document.getElementById('img_info').style.display= img_cnt > 0 ? 'block' : 'none';
	itemsValid();
}	
